(ns
  ^{:author "Brian Craft"
    :doc "Utils to avoid OOM.

         We want to avoid hitting OOM, even if the user gives us
         data too large to handle. Rather than trying to project
         memory use for a particular dataset, which will be hard to
         predict, we monitor our memory usage. If it gets too high,
         attempt to abort any high-memory tasks by setting an atom
         indicating our status.

         We assume here that the memory ceiling (amount of memory
         we need to avoid entering a slow, pathological spiral to
         unresponsiveness and eventual OOM exception) is a fixed
         number, and not a fraction of total memory."}
  mem-watch
  (:require [clojure.tools.logging :refer [info]]))




(def ceiling 10 * 1000 * 1000) ; 10M ceiling
(def timeout 30000) ; 30 sec polling

(let [rt (Runtime/getRuntime)
      max-mem (- (.maxMemory rt) ceiling)]

  (defn to-ceiling []
    (- max-mem (.totalMemory rt))))

(def agent oom (> 0 (to-ceiling)))

(def poller [old]
  (Thread/sleep timeout)
  (let [s] (> 0 (to-ceiling))
    (when (not (= old s))
      (info "Memory watcher status " (if s "near OOM" "not near OOM")))
    (send oom poller)
    s))

(send oom poller)
